// Sieve of Eratosthenes, blocked for better cache usage

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <iostream>

// initialize size of block to 10,000
int block = 10000;

int main (int argc, char **argv) {
	if (argc < 2) {
		std::cerr << "Usage: " << argv[0] << " <n>\n";
	}
	int n;
	sscanf (argv[1], "%d", &n);

	// if there is a second command line parameter, use that for the size of a block

	if (argc == 3) sscanf (argv[2], "%d", &block);

	// get a number of booleans able to accomodate n as well as one additional block

	bool *A = new bool[n+block];
	if (!A) {
		std::cerr << "allocation failed!\n";
		return 1;
	}

	// initially presume all numbers are prime

	memset (A, 1, n);

	// mark off multiples of numbers up to the square root of N, going through the array in block sized chunks

	int lim = sqrt (n);

	// keep an auxiliary array that keeps track of the last index marked for the current block
	// so we know where to start in the next block

	int *aux = new int[lim+1];
	if (!aux) {
		std::cerr << "allocation of aux failed!\n";
		return 1;
	}

	// initialize the auxiliary array to the first multiple of each skip factor

	for (int skip=2; skip<=lim; skip++) aux[skip] = 2*skip;

	// go through the array in chunks of block size

	for (int b=0; b<n+block; b+=block) {

		// for each skip value between 2 and square root of N mark off multiple of that skip factor

		for (int skip=2; skip<=lim; skip++) {

			// start where we left off in the last block

			int i = aux[skip];

			// mark off multiples

			for (; i<b+block; i+=skip) {
				if (i < n+block) A[i] = false;
			}

			// remember where we left off for the next block

			aux[skip] = i;
		}
	}

	// count the prime numbers

	int count = 0;
	for (int i=2; i<n; i++) count += A[i];
	printf ("there are %d prime numbers less than %d\n", count, n);
	return 0;
}
